/*
 * ICESOFT COMMERCIAL SOURCE CODE LICENSE V 1.1
 *
 * The contents of this file are subject to the ICEsoft Commercial Source
 * Code License Agreement V1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the
 * License at
 * http://www.icesoft.com/license/commercial-source-v1.1.html
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations under
 * the License.
 *
 * Copyright 2009-2013 ICEsoft Technologies Canada, Corp. All Rights Reserved.
 */

package com.icesoft.icepush;

import java.io.OutputStream;
import java.io.InputStream;
import java.net.Socket;
import java.security.KeyStore;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import org.icepush.NotificationProvider;
import org.icepush.OutOfBandNotifier;
import org.icepush.PushNotification;
import org.icepush.util.ExtensionRegistry;
import java.util.logging.Level;
import java.util.logging.Logger;

public class APNSNotificationProvider implements NotificationProvider {
    private static final Logger log = Logger.getLogger(APNSNotificationProvider.class.getName());
    // hardcoded as specified by Apple
    private String HOST = "gateway.push.apple.com";
    private int PORT = 2195;
    private static String KEYSTOREPATH = 
            "/WEB-INF/apns-keystore.p12";
    private static String KEYSTOREPASS = 
            "org.icemobile.apns.keyStorePassword";
    private SSLSocketFactory sslFactory;
    OutOfBandNotifier outOfBandNotifier;

    public APNSNotificationProvider(ServletContext servletContext) {

        InputStream keyStream = servletContext
                .getResourceAsStream(KEYSTOREPATH);
        if (null != keyStream)  {
            try {
                log.fine("Loading custom APNS keystore " + KEYSTOREPATH);
                String passString = servletContext
                    .getInitParameter(KEYSTOREPASS);
                passString = System.getProperty(KEYSTOREPASS, passString);
                char[] passPhrase = "".toCharArray();
                if (null != passString)  {
                    passPhrase = passString.toCharArray();
                }
                KeyManagerFactory keyManagerFactory = null;
                try {
                    keyManagerFactory = 
                            KeyManagerFactory.getInstance("SunX509");
                } catch (Exception e)  {
                    log.fine("No SunX509 KeyManagerFactory: " + e);
                }
                if (null == keyManagerFactory)  {
                    try {
                        keyManagerFactory = 
                                KeyManagerFactory.getInstance("IbmX509");
                    } catch (Exception e)  {
                        log.fine("No IbmX509 KeyManagerFactory: " + e);
                    }
                }

                KeyStore keyStore = KeyStore.getInstance("pkcs12");
                keyStore.load(keyStream, passPhrase);
                keyManagerFactory.init(keyStore, passPhrase);
                SSLContext sslContext = SSLContext.getInstance("TLS");
                sslContext.init(keyManagerFactory.getKeyManagers(), 
                    null, null);
                sslFactory = (SSLSocketFactory) 
                        sslContext.getSocketFactory();
            } catch (Exception e)  {
                log.log(Level.SEVERE, "Unable to initialize APNS keystore: ", e);
            }
        }
        if (null == sslFactory) {
            sslFactory = (SSLSocketFactory) SSLSocketFactory.getDefault();
        }
        log.info("APNS initialzed SSL " + sslFactory);
    }

    public void registerWith(OutOfBandNotifier outOfBandNotifier) {
        outOfBandNotifier.registerProvider("apns", this);
        this.outOfBandNotifier = outOfBandNotifier;
    }

    public void send(String uri, PushNotification notification) {
        log.info("APNSNotificationProvider.send " + uri);
        outOfBandNotifier.trace("APNSNotificationProvider.send " + uri);
        //Inefficient connect and close for now
        String hexToken = uri.substring("apns:".length());
        String body = notification.getSubject();
        byte[] apnsNotification = getEnhancedNotification(hexToken, 
            "{\"aps\":{\"alert\":\"" + body + "\"}}");
        try {
            SSLSocket apnsSocket = (SSLSocket) sslFactory.createSocket(HOST, PORT);
            OutputStream sout = apnsSocket.getOutputStream();
            sout.write(apnsNotification);
            sout.flush();
            sout.close();
        } catch (Exception e) {
            log.log(Level.SEVERE, "Failed to send APNS notification: ", e);
            outOfBandNotifier.trace("APNSNotificationProvider.send failed: " + e);
        }
    }

    public static byte[] getSimpleNotification(
            String hexToken, String payload )  {
        byte[] payloadBytes = payload.getBytes();
        int payloadLen = payloadBytes.length;
        //command + token len + token + payload len + payload
        byte[] notiBytes = new byte[1 + 2 + 32 + 2 + payloadBytes.length];
        int i = 0;
        //command
        notiBytes[i++] = 0;
        //token len
        notiBytes[i++] = 0;
        notiBytes[i++] = 32;
        //token
        for (int j = 0; j < hexToken.length(); j += 2)  {
            byte nextByte = (byte) Integer.parseInt(hexToken.substring(j, j + 2), 16);
            notiBytes[i++] = nextByte;
        }
        //payload len assumed 16 bit
        notiBytes[i++] = (byte) (payloadLen >> 8);
        notiBytes[i++] = (byte) (payloadLen & 0xff);
        //payload
        System.arraycopy(payloadBytes, 0, notiBytes, i, payloadLen);
        
        return notiBytes;
    }


    public static byte[] getEnhancedNotification(
            String hexToken, String payload )  {
        byte[] payloadBytes = payload.getBytes();
        int payloadLen = payloadBytes.length;
        //command + identifier + expiry + token len + token + payload len + payload
        byte[] notiBytes = new byte[1 + 4 + 4 + 2 + 32 + 2 + payloadBytes.length];
        int i = 0;
        //command
        notiBytes[i++] = 1;
        //identifier
        notiBytes[i++] = 0;
        notiBytes[i++] = 6;
        notiBytes[i++] = 6;
        notiBytes[i++] = 6;
        //expiry
        notiBytes[i++] = 0;
        notiBytes[i++] = 0;
        notiBytes[i++] = 0;
        notiBytes[i++] = 0;
        //token len
        notiBytes[i++] = 0;
        notiBytes[i++] = 32;
        //token
        for (int j = 0; j < hexToken.length(); j += 2)  {
            byte nextByte = (byte) Integer.parseInt(hexToken.substring(j, j + 2), 16);
            notiBytes[i++] = nextByte;
        }
        //payload len assumed 16 bit
        notiBytes[i++] = (byte) (payloadLen >> 8);
        notiBytes[i++] = (byte) (payloadLen & 0xff);
        //payload
        System.arraycopy(payloadBytes, 0, notiBytes, i, payloadLen);
        
        return notiBytes;
    }

    public static String toHexString(byte[] bytes)  {
        StringBuffer buf = new StringBuffer();
        for (int j = 0; j < bytes.length; j++)  {
            byte b = bytes[j];
            if ((b >= 0) && (b < 0x10))  {
                //print leading zero
                buf.append("0");
            }
            buf.append(Integer.toHexString((int) b & 0xff));
        }
        return buf.toString();
    }


    public static class AutoRegister implements ServletContextListener {
        public void contextInitialized(ServletContextEvent servletContextEvent) {
            ServletContext servletContext = servletContextEvent
                    .getServletContext();
            ExtensionRegistry.addExtension( servletContext, 10,
                    NotificationProvider.class.getName(),
                    new APNSNotificationProvider(servletContext) );
        }

        public void contextDestroyed(ServletContextEvent servletContextEvent) {
        }
    }
}
